from next_action.todotxt import Task
from next_action.pick_action import next_action_based_on_priority
from next_action.arguments import parse_arguments


def next_action() -> None:
    arguments = parse_arguments()
    filename: str = arguments.file
    try:
        todotxt_file = open(filename, "r")
    except FileNotFoundError:
        print("Can't find {0}".format(filename))
        return
    with todotxt_file:
        tasks = [Task(line.strip()) for line in todotxt_file.readlines()]
    action = next_action_based_on_priority(tasks, arguments.context)
    print(action.text if action else "Nothing to do!")



import argparse
import os
import shutil

import next_action


class ContextAction(argparse.Action):
    """ A context action that checks whether contexts start with an @. """
    def __call__(self, parser, namespace, values, option_string=None):
        if not values:
            return
        if values.startswith("@"):
            setattr(namespace, self.dest, values)
        else:
            parser.error("Contexts should start with an @.")


def parse_arguments() -> argparse.Namespace:
    """ Parse the command line arguments. """
    # Ensure that the help info is printed using all columns available
    os.environ['COLUMNS'] = str(shutil.get_terminal_size().columns)

    parser = argparse.ArgumentParser(description="Show the next action in your todo.txt",
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("--version", action="version", version="Next-action {0}".format(next_action.__version__))
    parser.add_argument("-f", "--file", help="filename of the todo.txt file to read",
                        type=str, default="todo.txt")
    parser.add_argument("context", metavar="@CONTEXT", help="Show the next action in the specified context", nargs="?",
                        type=str, action=ContextAction)
    return parser.parse_args()



import datetime
import re
from typing import Optional, Set


class Task(object):
    """ A task from a line in a todo.txt file. """
    def __init__(self, todo_txt: str) -> None:
        self.text = todo_txt

    def __repr__(self) -> str:
        return "{0}<{1}>".format(self.__class__.__name__, self.text)

    def contexts(self) -> Set[str]:
        """ Return the contexts of the task. """
        return self.__prefixed_items("@")

    def projects(self) -> Set[str]:
        """ Return the projects of the task. """
        return self.__prefixed_items(r"\+")

    def priority(self) -> Optional[str]:
        """ Return the priority of the task """
        match = re.match(r"\(([A-Z])\) ", self.text)
        return match.group(1) if match else None

    def creation_date(self) -> Optional[datetime.date]:
        """ Return the creation date of the task. """
        match = re.match(r"(?:\([A-Z]\) )?(\d\d\d\d)-(\d\d)-(\d\d)", self.text)
        if match:
            try:
                return datetime.date(*(int(group) for group in match.groups()))
            except ValueError:
                pass
        return None

    def is_completed(self) -> bool:
        """ Return whether the task is completed or not. """
        return self.text.startswith("x ")

    def __prefixed_items(self, prefix: str) -> Set[str]:
        """ Return the prefixed items in the task. """
        return {match.group(1) for match in re.finditer(" {0}([^ ]+)".format(prefix), self.text)}



import os
import sys
import unittest
from unittest.mock import patch, call

from next_action.arguments import parse_arguments


class ArgumentParserTest(unittest.TestCase):
    """ Unit tests for the argument parses. """

    @patch.object(sys, "argv", ["next_action"])
    def test_default_filename(self):
        """ Test that the argument parser has a default filename. """
        self.assertEqual("todo.txt", parse_arguments().file)

    @patch.object(sys, "argv", ["next_action", "-f", "my_todo.txt"])
    def test_filename_argument(self):
        """ Test that the argument parser accepts a filename. """
        self.assertEqual("my_todo.txt", parse_arguments().file)

    @patch.object(sys, "argv", ["next_action", "--file", "my_other_todo.txt"])
    def test_long_filename_argument(self):
        """ Test that the argument parser accepts a filename. """
        self.assertEqual("my_other_todo.txt", parse_arguments().file)

    @patch.object(sys, "argv", ["next_action"])
    def test_no_context(self):
        """ Test that the argument parser returns no context if the user doesn't pass one. """
        self.assertEqual(None, parse_arguments().context)

    @patch.object(sys, "argv", ["next_action", "@home"])
    def test_one_context(self):
        """ Test that the argument parser returns the context if the user passes one. """
        self.assertEqual("@home", parse_arguments().context)

    @patch.object(sys, "argv", ["next_action", "home"])
    @patch.object(sys.stderr, "write")
    def test_faulty_context(self, mock_stderr_write):
        """ Test that the argument parser exits if the context is faulty. """
        os.environ['COLUMNS'] = "120"  # Fake that the terminal is wide enough.
        self.assertRaises(SystemExit, parse_arguments)
        self.assertEqual([call("usage: next_action [-h] [--version] [-f FILE] [@CONTEXT]\n"),
                          call("next_action: error: Contexts should start with an @.\n")],
                         mock_stderr_write.call_args_list)


import unittest

from next_action import todotxt, pick_action


class PickActionTest(unittest.TestCase):
    """ Unit tests for the pick action algorithm. """

    def test_no_tasks(self):
        """ Test that no tasks means no next action. """
        self.assertEqual(None, pick_action.next_action_based_on_priority([]))

    def test_one_task(self):
        """ If there is one task, that one is the next action. """
        task = todotxt.Task("Todo")
        self.assertEqual(task, pick_action.next_action_based_on_priority([task]))

    def test_multiple_tasks(self):
        """ If there are multiple tasks, the first is the next action. """
        task1 = todotxt.Task("Todo 1")
        task2 = todotxt.Task("Todo 2")
        self.assertEqual(task1, pick_action.next_action_based_on_priority([task1, task2]))

    def test_higher_prio_goes_first(self):
        """ If there are multiple tasks with different priorities, the task with the
            highest priority is the next action. """
        task1 = todotxt.Task("Todo 1")
        task2 = todotxt.Task("(B) Todo 2")
        task3 = todotxt.Task("(A) Todo 3")
        self.assertEqual(task3, pick_action.next_action_based_on_priority([task1, task2, task3]))

    def test_completed_tasks_are_not_next_action_based_on_priority(self):
        """ If all tasks are completed, there's no next action. """
        completed_task1 = todotxt.Task("x Completed")
        completed_task2 = todotxt.Task("x Completed too")
        self.assertEqual(None, pick_action.next_action_based_on_priority([completed_task1, completed_task2]))

    def test_completed_task_is_not_next_action_based_on_priority(self):
        """ If there's one completed and one uncompleted task, the uncompleted one is the next action. """
        completed_task = todotxt.Task("x Completed")
        uncompleted_task = todotxt.Task("Todo")
        self.assertEqual(uncompleted_task,
                         pick_action.next_action_based_on_priority([completed_task, uncompleted_task]))

    def test_next_action_limited_to_context(self):
        """ Test that the next action can be limited to a specific context. """
        task1 = todotxt.Task("Todo 1 @work")
        task2 = todotxt.Task("(B) Todo 2 @work")
        task3 = todotxt.Task("(A) Todo 3 @home")
        self.assertEqual(task2, pick_action.next_action_based_on_priority([task1, task2, task3], context="work"))



from typing import Optional, Iterable

from .todotxt import Task


def next_action_based_on_priority(tasks: Iterable[Task], context: str = "") -> Optional[Task]:
    """ Return the next action from the collection of tasks. """
    uncompleted_tasks = [task for task in tasks if not task.is_completed()]
    if context:
        tasks_in_context = [task for task in tasks if context.strip("@") in task.contexts()]
    else:
        tasks_in_context = uncompleted_tasks
    sorted_tasks = sorted(tasks_in_context, key=lambda task: task.priority() or "ZZZ")
    return sorted_tasks[0] if sorted_tasks else None



import os
import sys
import unittest
from unittest.mock import patch, mock_open, call

from next_action.cli import next_action
from next_action import __version__


class CLITest(unittest.TestCase):
    """ Unit tests for the command-line interface. """

    @patch.object(sys, "argv", ["next_action"])
    @patch("next_action.cli.open", mock_open(read_data=""))
    @patch.object(sys.stdout, "write")
    def test_empty_task_file(self, mock_stdout_write):
        """ Test the response when the task file is empty. """
        next_action()
        self.assertEqual([call("Nothing to do!"), call("\n")], mock_stdout_write.call_args_list)

    @patch.object(sys, "argv", ["next_action"])
    @patch("next_action.cli.open", mock_open(read_data="Todo\n"))
    @patch.object(sys.stdout, "write")
    def test_one_task(self, mock_stdout_write):
        """ Test the response when the task file has one task. """
        next_action()
        self.assertEqual([call("Todo"), call("\n")], mock_stdout_write.call_args_list)

    @patch.object(sys, "argv", ["next_action"])
    @patch("next_action.cli.open")
    @patch.object(sys.stdout, "write")
    def test_missing_file(self, mock_stdout_write, mock_open):
        """ Test the response when the task file can't be found. """
        mock_open.side_effect = FileNotFoundError
        next_action()
        self.assertEqual([call("Can't find todo.txt"), call("\n")], mock_stdout_write.call_args_list)

    @patch.object(sys, "argv", ["next_action", "--help"])
    @patch.object(sys.stdout, "write")
    def test_help(self, mock_stdout_write):
        """ Test the help message. """
        os.environ['COLUMNS'] = "120"  # Fake that the terminal is wide enough.
        self.assertRaises(SystemExit, next_action)
        self.assertEqual(call("""usage: next_action [-h] [--version] [-f FILE] [@CONTEXT]

Show the next action in your todo.txt

positional arguments:
  @CONTEXT              Show the next action in the specified context (default: None)

optional arguments:
  -h, --help            show this help message and exit
  --version             show program's version number and exit
  -f FILE, --file FILE  filename of the todo.txt file to read (default: todo.txt)
"""),
                        mock_stdout_write.call_args_list[0])

    @patch.object(sys, "argv", ["next_action", "--version"])
    @patch.object(sys.stdout, "write")
    def test_version(self, mock_stdout_write):
        """ Test that --version shows the version number. """
        self.assertRaises(SystemExit, next_action)
        self.assertEqual([call("Next-action {0}\n".format(__version__))], mock_stdout_write.call_args_list)